/*-
 * Copyright Sergey Kosyakov ks@itp.ac.ru 1999
 * Copyright (C)2002..2020 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ident "@(#)Copyright Sergey Kosyakov ks@itp.ac.ru 1999"
#ident "@(#)Copyright (C)2002..2020 @BABOLO http://www.babolo.ru/"
#ident "$Id: alarm.c,v 1.16 2020/11/27 23:57:27 babolo Exp $"

#define BLIN_COMPAT  3

#include <sys/types.h>
#include <sys/time.h>
#include <sysexits.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <netinet/in.h>
#include <openssl/rsa.h>
#include <babolo/BLINflag.h>
#include "tund.h"

#define	TIMER_GRAN	10	                                    /* alarm timer granularity, sec. */

static int        has_exp_oper = 0;
extern tund_conf *tund_cnf;

static void
h_sigalarm(int sig, int code, struct sigcontext *sc) {
#   define blin_internal_flags (tund_cnf->flags)
    tund_tunnel *tun;
    int i;

    ifBLIN_QX5("h_sigalarm(%d %d 0x%x)", sig, code, sc);
    if  (!has_exp_oper) return;
    ifBLIN_QX3("Has %d expired operations", has_exp_oper);
    has_exp_oper = 0;
    for (i = 0; i < tund_cnf->stun->length; i++) {
        tun = (tund_tunnel *)tund_cnf->stun->buf[i];
        if  (tun->f_exp) {
            has_exp_oper++;
            if  ((--(tun->timer)) <= 0) {
                ifBLIN_QX5("Call %"BLIN_X" on tunnel #%d", BLIN_I(tun->f_exp), i);
                tun->f_exp(tund_cnf, i);
                has_exp_oper--;
    }   }   }
#   undef blin_internal_flags
}

int
tund_init_alarm(tund_conf *cnf) {
#   define blin_internal_flags (cnf->flags)
    struct itimerval ival;
    struct sigaction sa;
    int ex = EX_OK;

    sa.sa_handler = h_sigalarm;
    sigemptyset(&sa.sa_mask);
    sigaddset(&sa.sa_mask, SIGALRM);
    sa.sa_flags = SA_RESTART;
    if  (!!sigaction(SIGALRM, &sa, NULL)) {
        ifBLIN_QW0("sigaction(SIGALRM)");
        ex = EX_OSERR;
        goto out;
    }
    ival.it_interval.tv_sec = TIMER_GRAN;
    ival.it_interval.tv_usec = 0;
    ival.it_value.tv_sec = TIMER_GRAN;
    ival.it_value.tv_usec = 0;
    if  (!!setitimer(ITIMER_REAL, &ival, NULL)) {
        ifBLIN_QX0("setitimer(ITIMER_REAL)");
        ex = EX_OSERR;
        goto out;
    }
out:
    return(ex);
#   undef blin_internal_flags
}

int
tund_add_expired_operation(tund_conf *cnf, int tun_ind, int t, int (*f_exp)(tund_conf *cnf, int)) {
#   define blin_internal_flags (cnf->flags)
    tund_tunnel *tun = (tund_tunnel *)cnf->stun->buf[tun_ind];
    int ex = EX_OK;
    sigset_t sset;

    sigemptyset(&sset);
    sigaddset(&sset, SIGALRM);
    sigprocmask(SIG_BLOCK, &sset, NULL);
    tun->timer = t / TIMER_GRAN;
    tun->f_exp = f_exp;
    has_exp_oper++;
    ifBLIN_QX3("Set up expiration callback 0x%x for %d sec on tunnel %u", f_exp, t, tun->tunum);
    sigprocmask(SIG_UNBLOCK, &sset, NULL);
    return(ex);
#   undef blin_internal_flags
}

int
tund_remove_expired_operation(tund_conf *cnf, int tun_ind) {
#   define blin_internal_flags (cnf->flags)
    tund_tunnel *tun = (tund_tunnel *)cnf->stun->buf[tun_ind];
    int ex = EX_OK;
    sigset_t sset;

    sigemptyset(&sset);
    sigaddset(&sset, SIGALRM);
    sigprocmask(SIG_BLOCK, &sset, NULL);
    if  (!tun->f_exp) goto L_end;
    tun->timer = 0;
    tun->f_exp = NULL;
    has_exp_oper--;
    ifBLIN_QX3("Remove expiration callback tunnel %u", tun->tunum);
L_end:
    sigprocmask(SIG_UNBLOCK, &sset, NULL);
    return(ex);
#   undef blin_internal_flags
}
