/*-
 * Copyright Sergey Kosyakov ks@itp.ac.ru 1999
 * Copyright (C)2002..2020 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ident "@(#)Copyright Sergey Kosyakov ks@itp.ac.ru 1999"
#ident "@(#)Copyright (C)2002..2020 @BABOLO http://www.babolo.ru/"
#ident "$Id: scb.c,v 1.22 2020/12/05 17:03:19 babolo Exp $"

#define BLIN_COMPAT   3
#define RECOBE_COMPAT 4

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <arpa/inet.h>
#include <sysexits.h>
#include <stdio.h>
#include <errno.h>
#include <fcntl.h>
#include <netdb.h>
#include <openssl/rsa.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <netinet/sctp.h>
#include <babolo/BLINflag.h>
#include <babolo/recobe.h>
#include "tund.h"

u_char packetBuf[TUND_PKT_MAXSZ];

static int
ipdump(tund_conf *cnf, struct ip *pip, size_t len) {
#   define blin_internal_flags (cnf->flags)
    struct protoent *proto;
    int              ex = EX_OK;
    ssize_t          hlen;
    u_int32_t        src;
    u_int32_t        dst;

    proto = getprotobynumber(pip->ip_p);
    src = recobe_load((const void *)&pip->ip_src.s_addr, 4);
    dst = recobe_load((const void *)&pip->ip_dst.s_addr, 4);
    hlen = pip->ip_hl << 2;
    ifBLIN_QW1( "v%u[%u][%u=%u]tos=%02X id=%04X off=%04X ttl=%u"
              , pip->ip_v
              , hlen
              , ntohs(pip->ip_len)
              , len
              , pip->ip_tos
              , ntohs(pip->ip_id)
              , ntohs(pip->ip_off)
              , pip->ip_ttl
              );
    if  (  pip->ip_p == IPPROTO_TCP
        && (ntohs(pip->ip_off) & IP_OFFMASK) == 0
        && ntohs(pip->ip_len) == len
        && hlen < len
        && len - hlen >= sizeof(struct tcphdr)
        ) {
        struct tcphdr *tc = (struct tcphdr *)((char*)pip + hlen);

        ifBLIN_QX1( "%s %u.%u.%u.%u:%u > %u.%u.%u.%u:%u%s%s%s%s%s%s%s%s"
                  , proto->p_name
                  , src & 0x0FF
                  , (src >> 8) & 0x0FF
                  , (src >> 16) & 0x0FF
                  , (src >> 24) & 0x0FF
                  , ntohs(tc->th_sport)
                  , dst & 0x0FF
                  , (dst >> 8) & 0x0FF
                  , (dst >> 16) & 0x0FF
                  , (dst >> 24) & 0x0FF
                  , ntohs(tc->th_dport)
                  , (tc->th_flags & TH_FIN) ? " fin" : ""
                  , (tc->th_flags & TH_SYN) ? " syn" : ""
                  , (tc->th_flags & TH_RST) ? " rst" : ""
                  , (tc->th_flags & TH_PUSH)? " push": ""
                  , (tc->th_flags & TH_ACK) ? " ack" : ""
                  , (tc->th_flags & TH_URG) ? " urg" : ""
                  , (tc->th_flags & TH_ECE) ? " ece" : ""
                  , (tc->th_flags & TH_CWR) ? " cwr" : ""
                  );
    } else if (  pip->ip_p == IPPROTO_UDP
              && (ntohs(pip->ip_off) & IP_OFFMASK) == 0
              && ntohs(pip->ip_len) == len
              && hlen < len
              && len - hlen >= sizeof(struct udphdr)
              ) {
        struct udphdr *uc = (struct udphdr *)((char*)pip + hlen);

        ifBLIN_QX1( "%s %u.%u.%u.%u:%u > %u.%u.%u.%u:%u"
                  , proto->p_name
                  , src & 0x0FF
                  , (src >> 8) & 0x0FF
                  , (src >> 16) & 0x0FF
                  , (src >> 24) & 0x0FF
                  , ntohs(uc->uh_sport)
                  , dst & 0x0FF
                  , (dst >> 8) & 0x0FF
                  , (dst >> 16) & 0x0FF
                  , (dst >> 24) & 0x0FF
                  , ntohs(uc->uh_dport)
                  );
    } else if (  pip->ip_p == IPPROTO_SCTP
              && (ntohs(pip->ip_off) & IP_OFFMASK) == 0
              && ntohs(pip->ip_len) == len
              && hlen < len
              && len - hlen >= sizeof(struct sctphdr)
              ) {
        struct sctphdr *sc = (struct sctphdr *)((char*)pip + hlen);
        u_int32_t       v_tag = recobe_load((const void *)&sc->v_tag, 4);

        ifBLIN_QX1( "%s %u.%u.%u.%u:%u > %u.%u.%u.%u:%u v%08X"
                  , proto->p_name
                  , src & 0x0FF
                  , (src >> 8) & 0x0FF
                  , (src >> 16) & 0x0FF
                  , (src >> 24) & 0x0FF
                  , ntohs(sc->src_port)
                  , dst & 0x0FF
                  , (dst >> 8) & 0x0FF
                  , (dst >> 16) & 0x0FF
                  , (dst >> 24) & 0x0FF
                  , ntohs(sc->dest_port)
                  , ntohl(v_tag)
                  );
    } else {
        ifBLIN_QX1( "[%u]%s %u.%u.%u.%u > %u.%u.%u.%u"
                  , pip->ip_p
                  , proto ? proto->p_name : ""
                  , src & 0x0FF
                  , (src >> 8) & 0x0FF
                  , (src >> 16) & 0x0FF
                  , (src >> 24) & 0x0FF
                  , dst & 0x0FF
                  , (dst >> 8) & 0x0FF
                  , (dst >> 16) & 0x0FF
                  , (dst >> 24) & 0x0FF
                  );
    }
    return(ex);
#   undef blin_internal_flags
}

static int
ks_recvfrom( tund_conf       *cnf
           , int              s
           , void            *buf
           , size_t           len
           , int              flags
           , struct sockaddr *from
           , socklen_t       *fromlen
           ) {
#   define blin_internal_flags (cnf->flags)
    struct sockaddr_in *sain;
    int                 ret;

    if  (0 > (ret = recvfrom(s, buf, len, flags, from, fromlen))) {
        ifBLIN_QW0("Error in ks_recvfrom(s=%d) while recvfrom", s);
    } else {
        sain = (struct sockaddr_in *)from;
        ifBLIN_QX6( "Recvfrom s=%d len=%d %d[%d]%u.%u.%u.%u:%u?%u"
                  , s
                  , ret
                  , from->sa_family
                  , from->sa_len
                  , sain->sin_addr.s_addr & 0x00FF
                  , (sain->sin_addr.s_addr >> 8) & 0x00FF
                  , (sain->sin_addr.s_addr >> 16) & 0x00FF
                  , (sain->sin_addr.s_addr >> 24) & 0x00FF
                  , ntohs(sain->sin_port)
                  , sain->sin_port
                  );
    }
    return(ret);
#   undef blin_internal_flags
}

int
tund_ks_send( tund_conf       *cnf
            , int              tun_ind
            , int              s
            , void            *msg
            , size_t           len
            , struct sockaddr *to
            , int              tolen
            ) {
#   define blin_internal_flags (cnf->flags)
    struct sockaddr_in *sain = (struct sockaddr_in *)to;
    int                 ex = EX_OK;
    int                 ret;

    ret = sendto(s, msg, len, 0, to, tolen);
    if  (ret < 0) {
        if  (errno == EINVAL) {
            ifBLIN_QX2("Dropped illegal IP packet (t=%d so=%d l=%d)", tun_ind, s, len);
            ex = EX_USAGE;
            goto out;
        } else {
            ifBLIN_QW2( "sendto(t=%d s=%d) %d[%d]%u.%u.%u.%u:%u?%u"
                      , tun_ind
                      , s
                      , to->sa_family
                      , to->sa_len
                      , (sain->sin_addr.s_addr >> 24) & 0x00FF
                      , (sain->sin_addr.s_addr >> 16) & 0x00FF
                      , (sain->sin_addr.s_addr >> 8) & 0x00FF
                      , sain->sin_addr.s_addr & 0x00FF
                      , ntohs(sain->sin_port)
                      , sain->sin_port
                      );
            ex = EX_NOPERM;
            goto out;
    }   }
    if  (ret != len) {
        ifBLIN_QX0( "sendto(t=%d s=%d): %d<>%d %d[%d]%u.%u.%u.%u:%u?%u"
                  , tun_ind
                  , s
                  , ret
                  , len
                  , to->sa_family
                  , to->sa_len
                  , (sain->sin_addr.s_addr >> 24) & 0x00FF
                  , (sain->sin_addr.s_addr >> 16) & 0x00FF
                  , (sain->sin_addr.s_addr >> 8) & 0x00FF
                  , sain->sin_addr.s_addr & 0x00FF
                  , ntohs(sain->sin_port)
                  , sain->sin_port
                  );
        errno = EIO;
        ex = EX_IOERR;
        goto out;
    }
    ifBLIN_QX6( "Sended %d bytes to(t=%d s=%d): %d[%d]%u.%u.%u.%u:%u?%u"
              , ret
              , tun_ind
              , s
              , to->sa_family
              , to->sa_len
              , (sain->sin_addr.s_addr >> 24) & 0x00FF
              , (sain->sin_addr.s_addr >> 16) & 0x00FF
              , (sain->sin_addr.s_addr >> 8) & 0x00FF
              , sain->sin_addr.s_addr & 0x00FF
              , ntohs(sain->sin_port)
              , sain->sin_port
              );
out:
    return(ex);
#   undef blin_internal_flags
}

static int
input_control(tund_conf *cnf, int tun_ind, int buf_len) {
#   define blin_internal_flags (cnf->flags)
    int          ex = EX_OK;
    tund_tunnel *tun;

    tun = (tund_tunnel *)cnf->stun->buf[tun_ind];
    ifBLIN_QX5("Received %d bytes from control of tunnel %u", buf_len, tun->tunum);
    if  (!!(ex = tund_proc_cp(cnf, tun_ind, buf_len))) {
        ifBLIN_QW0("tund_proc_cp");
        goto out;
    }
out:
    return(ex);
#   undef blin_internal_flags
}

int
tund_read_pkt(tund_conf *cnf, int tun_ind, int s) {
#   define blin_internal_flags (cnf->flags)
    int                ex = EX_OK;
    socklen_t          fromlen;
    int                buf_len;
    tund_pkthead      *phead;
    struct sockaddr_in from;

    fromlen = sizeof(from);
    buf_len = ks_recvfrom(cnf, s, packetBuf, TUND_PKT_MAXSZ, 0, (struct sockaddr *)&from, &fromlen);
    if  (0 > buf_len) {
        ifBLIN_QW0("ks_recvfrom");
        ex = EX_IOERR;
        goto out;
    }
    phead = (tund_pkthead *)packetBuf;
    if  ((buf_len < (sizeof(tund_pkthead) + 1)) || (phead->ver != TUND_PROTO_VER)) {
        ifBLIN_QX2( "Dropped packet (illegal packet) l=%d from %s:%d lh4=0x%x f=0x%x v=%d"
                  , buf_len
                  , inet_ntoa(from.sin_addr)
                  , (int)ntohs(from.sin_port)
                  , phead->tunum
                  , phead->flags
                  , phead->ver
                  );
        return(ex);
    }
    tun_ind = cnf->tunTable[phead->tunum] - 1;
    if  (tun_ind < 0) {
        ifBLIN_QX2( "Dropped packet (unknown tunnel) l=%d from %s:%d lh4=0x%x f=0x%x v=%d"
                  , buf_len
                  , inet_ntoa(from.sin_addr)
                  , (int)ntohs(from.sin_port)
                  , phead->tunum
                  , phead->flags
                  , phead->ver
                  );
        return(ex);
    }
    switch(phead->flags & 0x0003) {
    case TUND_PPF_DATA:
        {   int            body_len;
            tund_EDP      *en_data;
            tund_PDP      *data;
            u_char        *body;
            tund_tunnel   *tun;

            tun = (tund_tunnel *)cnf->stun->buf[tun_ind];
            if  (tun->flags & (TUND_CIPHER_MASK | TUND_F_RPWD)) {
                en_data = (tund_EDP *)packetBuf;
                tund_data_dec( cnf
                             , packetBuf + sizeof(tund_pkthead)
                             , buf_len-sizeof(tund_pkthead)
                             , tun_ind
                             );
                body = en_data->body;
                body_len = en_data->body_len;
            } else {
                data = (tund_PDP *)packetBuf;
                body_len = buf_len - sizeof(tund_pkthead);
                body = data->body;
            }
            ifBLIN_QX5("Received %d bytes from tunnel %u", body_len, tun->tunum);
            if  (tun->state & TUND_S_UP) {
                if  (!!tund_ks_send( cnf
                                   , tun_ind
                                   , tun->socketDivert
                                   , body
                                   , body_len
                                   , (struct sockaddr *)&(tun->l_addr)
                                   , sizeof(struct sockaddr_in)
                    )              ) {
                    (void)ipdump(cnf, (struct ip *)body, body_len);
                    if  (EIO == errno) exit(EX_IOERR);
                }
            } else {
                ifBLIN_QX5("Dropped %d bytes on tunnel %u", body_len, tun->tunum);
        }   }
        break;
    case TUND_PPF_CNT:
        if  (!!(ex = input_control(cnf, tun_ind, buf_len))) {
            ifBLIN_QW0("input_control");
            goto out;
        }
        break;
    default:
        ifBLIN_QX2( "Dropped packet len=%d from %s:%d lh4=0x%x flags=0x%x"
                  , buf_len
                  , inet_ntoa(from.sin_addr)
                  , (int)ntohs(from.sin_port)
                  , phead->tunum
                  , phead->flags
                  );
        return(ex);
    }
out:
    return(ex);
#   undef blin_internal_flags
}

int
tund_inp_dvrt(tund_conf *cnf, int tun_ind, int s) {
#   define blin_internal_flags (cnf->flags)
    int             ex = EX_OK;
    struct sockaddr from;
    socklen_t       fromlen = sizeof(from);
    tund_EDP       *en_data;
    int             buf_len;
    tund_PDP       *data;
    tund_tunnel    *tun;
    int             len;
    struct ip      *pip;

    tun = (tund_tunnel *)cnf->stun->buf[tun_ind];
    ifBLIN_QX5("Input packet on divert socket %d", s);
    if  (TUND_CIPHER_MASK & tun->flags) {
        en_data = (tund_EDP *)packetBuf;
        if  (0 > (len = ks_recvfrom(cnf, s, en_data->body, TUND_BUFSIZE, 0, &from, &fromlen))) {
            ifBLIN_QW0("ks_recvfrom enc");
            ex = EX_IOERR;
            goto out;
        }
        en_data->h.tunum = tun->tunum;
        en_data->h.flags = TUND_PPF_DATA;
        en_data->h.ver = TUND_PROTO_VER;
        en_data->rn = tund_get_rand(cnf);
        en_data->body_len = (u_int16_t)len;
        if  (0 > (buf_len = tund_data_enc(cnf, packetBuf + sizeof(tund_pkthead), len + 4, tun_ind))) {
            ifBLIN_QX2("No valid encription");
            goto out;
        }
        buf_len += sizeof(tund_pkthead);
        pip = (struct ip *)en_data->body;
    } else {
        data = (tund_PDP *)packetBuf;
        if  (0 > (len = ks_recvfrom(cnf, s, data->body, TUND_BUFSIZE, 0, &from, &fromlen))) {
            ifBLIN_QW0("ks_recvfrom");
            ex = EX_IOERR;
            goto out;
        }
        buf_len = len + sizeof(tund_pkthead);
        data->h.tunum = tun->tunum;
        data->h.flags = TUND_PPF_DATA;
        data->h.ver = TUND_PROTO_VER;
        pip = (struct ip *)data->body;
    }
    if  (tun->state & TUND_S_UP) {
        if  (!!tund_ks_send( cnf
                           , tun_ind
                           , cnf->isock
                           , packetBuf
                           , buf_len
                           , (struct sockaddr *)(&(tun->to))
                           , sizeof(struct sockaddr_in)
            )              ) {
            (void)ipdump(cnf, pip, len);
            if  (EIO == errno) exit(EX_IOERR);
        }
    } else {
        ifBLIN_QX5("Dropped %d bytes on tunnel %u", len, tun->tunum);
    }
out:
    return(ex);
#   undef blin_internal_flags
}