/*-
 * Copyright Sergey Kosyakov ks@itp.ac.ru 1999
 * Copyright (C)2002..2020 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ident "@(#)Copyright Sergey Kosyakov ks@itp.ac.ru 1999"
#ident "@(#)Copyright (C)2002..2020 @BABOLO http://www.babolo.ru/"
#ident "$Id: utils.c,v 1.20 2020/12/05 17:03:19 babolo Exp $"

#define BLIN_COMPAT  3

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/file.h>
#include <sysexits.h>
#include <stdlib.h>
#include <limits.h>
#include <unistd.h>
#include <string.h>
#include <syslog.h>
#include <stdarg.h>
#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <netdb.h>
#include <openssl/rsa.h>
#include <babolo/BLINflag.h>
#include "tund.h"

#define SEQ_INISIZE	16

/*****************************************************************************/
/* Sequence support                                                          */
/*****************************************************************************/

tund_seq *
tund_seq_new(tund_conf *cnf) {
#   define blin_internal_flags (cnf->flags)
    tund_seq *ret;

    if  (!(ret = calloc(1, sizeof(tund_seq)))) {
        ifBLIN_QW0("no mem for tund_seq");
    } else {
        ret->maxlen = SEQ_INISIZE;
        ret->length = 0;
        if  (!(ret->buf = calloc(SEQ_INISIZE, sizeof(void *)))) {
            ifBLIN_QW0("no mem for tund_seq->buf");
            free(ret);
            ret = NULL;
            errno = ENOMEM;
    }   }
    return(ret);
#   undef blin_internal_flags
}

int
tund_seq_append(tund_conf *cnf, tund_seq *s, void *item) {
#   define blin_internal_flags (cnf->flags)
    int    ex = EX_OK;
    void **buf;
    int    len;

    s->length++;
    if  (s->length >= s->maxlen) {
        len = s->maxlen * 2;
        if  (!(buf = calloc(len, sizeof(void *)))) {
            ifBLIN_QW0("no %lu mem for stun double", len * sizeof(void *));
            ex = -EX_OSERR;
            goto out;
        }
        bcopy(s->buf, buf, s->maxlen * sizeof(void *));
        free(s->buf);
        s->buf = buf;
        s->maxlen = len;
    }
    s->buf[s->length - 1] = item;
out:
    return(!ex ? (s->length - 1) : ex);
#   undef blin_internal_flags
}

/*****************************************************************************/
/* RNG                                                                       */
/*****************************************************************************/

static int32_t   *rng_F;
static int32_t    rng_S;
static int32_t    rng_RL  = 9689;
static int32_t    rng_D   = 471;
static u_int32_t *rng_R;
static u_int32_t  rng_msk = 0x7FFFFFFF;

int
tund_get_rand(tund_conf *cnf) {
#   define blin_internal_flags (cnf->flags)
    int tmp;
    int n;

    tmp = rng_R[*rng_F] ^ rng_R[rng_S];
    tmp &= rng_msk;
    rng_R[*rng_F] = tmp;
    (*rng_F)++;
    rng_S++;
    if  (*rng_F != rng_RL) {
        if  (rng_S == rng_RL) rng_S = 0;
    } else {
        *rng_F = 0;
    }
    if  (TUND_FOREG & cnf->flags) {
        n = *rng_F - rng_D;
        if  (n < 0) n = n + rng_RL;
        if  (n != rng_S) {
            ifBLIN_QX0("n=%d, rng_S=%d", n, rng_S);
            exit(1);
    }   }
    return(tmp);
#   undef blin_internal_flags
}

int
tund_init_rng(tund_conf *cnf) {
#   define blin_internal_flags (cnf->flags)
    int         ex = EX_OK;
    int         fd;
    struct stat sb;
    int         ret;

    if  (0 > (fd = open(cnf->seed, O_RDWR))) {
        ifBLIN_QW0("Can not open file %s", cnf->seed);
        ex = EX_IOERR;
        goto out;
    }
    if  (0 > (ret = fstat(fd, &sb))) {
        ifBLIN_QW0("Can not get stat of seed file");
        ex = EX_IOERR;
        goto out;
    }
    if  (sb.st_size != rng_RL * 4 + 4) {
        ifBLIN_QW0("Illegal size %d of seed file", (int)sb.st_size);
        ex = EX_DATAERR;
        goto out;
    } 
    rng_F = mmap(NULL, rng_RL * 4 + 4, PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
    if  (MAP_FAILED == rng_F) {
        ifBLIN_QW0("Can not mmap seed file");
        ex = EX_IOERR;
        goto out;
    }
    if  (*rng_F >= rng_RL || *rng_F < 0) {
        ifBLIN_QW0("Corrupted RNG seed file");
        ex = EX_DATAERR;
        goto out;
    }
    rng_R = (u_int32_t *)rng_F + 1;
    rng_S = *rng_F - rng_D;
    if  (rng_S < 0) rng_S = rng_RL + rng_S;
    ifBLIN_QX2("F=%d, S=%d\n", *rng_F, rng_S);
    /* for (ret = 0; ret < 200000; ret++) printf("%d\n", tund_get_rand()); */
out:
    return(ex);
#   undef blin_internal_flags
}
